/*
 * plot3d.c
 *
 *  Created on: 4 de abr de 2017
 *      Author: Alexandre
 */


/*
 * plot3d.c
 *
 *  Created on: 4 de mar de 2017
 *      Author: Alexandre
 *
 * Implements a 3d graphics pipeline.
 * This fast and simple engine is only capable of projecting isolated points.
 */

#include "plot3d.h"

// Constants
//
#define PARALLEL    0
#define PERSPECTIVE 1

#define ROUNDOFF  	1e-15
#define VERYLARGE 	1e+15

// Transformation matrix
//
double m00, m01, m02;
double m10, m11, m12;
double m20, m21, m22;
double m30, m31, m32;
//
// 3d fields
//
Vector3d center;	      	// center of projection
Vector3d pc;         	 	// transformed center of projection
Vector3d pen;        		// plotting pen

double distance;        	// view distance
double frontDistance;		// clipping plane
double backDistance;		// clipping plane
int ucs;					// UCS icon flag
//
// Window
//
double wxl, wyl;
double wxh, wyh;
double vxl, vyl;
double vxh, vyh;
double wsx, wsy;
//
int mode;
int clipping;
double scale = 0.8;
int background = BLACKBG;

DWORD X, Y, Z;
DWORD* pxls;
DWORD R, G, B;

BOOL showAxes = true, showGrid;

Vector3d position;      	// view reference point
Vector3d direction;  		// camera axis
Vector3d attitude;   		// view-up direction

DWORD WinColor(int r, int g, int b)
{
	DWORD dwStuff = ((r << 16) | (g << 8) | (b));
	return dwStuff;
}

void initPalette()
{
	R = WinColor(255, 0, 0);
	G = WinColor(0, 255, 0);
	B = WinColor(80, 80, 255);
	//
	X = WinColor(120, 20, 20);
	Y = WinColor(0, 100, 0);
	Z = WinColor(30, 30, 150);
}

void newTransform3()
{
	m00 = 1;
	m01 = 0;
	m02 = 0;
	//
	m10 = 0;
	m11 = 1;
	m12 = 0;
	//
	m20 = 0;
	m21 = 0;
	m22 = 1;
	//
	m30 = 0;
	m31 = 0;
	m32 = 0;
}

void translate(double tx, double ty, double tz)
{
	m30 += tx;
	m31 += ty;
	m32 += tz;
}

void rotateX(double s, double c)
{
	double t;
	t = m01 * c - m02 * s;
	m02 = m01 * s + m02 * c;
	m01 = t;
	//
	t = m11 * c - m12 * s;
	m12 = m11 * s + m12 * c;
	m11 = t;
	//
	t = m21 * c - m22 * s;
	m22 = m21 * s + m22 * c;
	m21 = t;
	//
	t = m31 * c - m32 * s;
	m32 = m31 * s + m32 * c;
	m31 = t;
}

void rotateY(double s, double c)
{
	double t;
	t = m00 * c + m02 * s;
	m02 = -m00 * s + m02 * c;
	m00 = t;
	//
	t = m10 * c + m12 * s;
	m12 = -m10 * s + m12 * c;
	m10 = t;
	//
	t = m20 * c + m22 * s;
	m22 = -m20 * s + m22 * c;
	m20 = t;
	//
	t = m30 * c + m32 * s;
	m32 = -m30 * s + m32 * c;
	m30 = t;
}

void rotateZ(double s, double c)
{
	double t;
	t = m00 * c - m01 * s;
	m01 = m00 * s + m01 * c;
	m00 = t;
	//
	t = m10 * c - m11 * s;
	m11 = m10 * s + m11 * c;
	m10 = t;
	//
	t = m20 * c - m21 * s;
	m21 = m20 * s + m21 * c;
	m20 = t;
	//
	t = m30 * c - m31 * s;
	m31 = m30 * s + m31 * c;
	m30 = t;
}

void viewPlaneTransform(Vector3d *p)
{
	double x = p->x * m00 + p->y * m10 + p->z * m20 + m30;
	double y = p->x * m01 + p->y * m11 + p->z * m21 + m31;
	double z = p->x * m02 + p->y * m12 + p->z * m22 + m32;
	p->x = x;
	p->y = y;
	p->z = z;
}

void makePerspectiveTransformation()
{
	pc.x = center.x;
	pc.y = center.y;
	pc.z = center.z;
	viewPlaneTransform(&pc);
	if(pc.z < 0)
		perror("Center of Projection behind View Plane.");
}

void makeViewPlaneTransformation()
{
	// Start with the identity matrix
	//
	newTransform3();
	//
	// Translate so that view plane center is new origin
	//
	translate(-(position.x + direction.x * distance),
				  -(position.y + direction.y * distance),
			  -(position.z + direction.z * distance));
	//
	// Rotate so that view plane normal is z axis
	//
	double v = sqrt(direction.y * direction.y + direction.z * direction.z);
	if(v > ROUNDOFF)
		rotateX(-direction.y / v, -direction.z / v);
	rotateY(direction.x, v);
	//
	// Determine the view-up direction in these new coordinates
	//
	double xup_vp = attitude.x * m00 + attitude.y * m10 + attitude.z * m20;
	double yup_vp = attitude.x * m01 + attitude.y * m11 + attitude.z * m21;
	//
	// Determine rotation needed to make view-up vertical
	//
	double rup = sqrt(xup_vp * xup_vp + yup_vp * yup_vp);
	if(rup < ROUNDOFF)
		perror("set-view-up along view-plane normal");
	rotateZ(xup_vp / rup, yup_vp / rup);
	//
	// Transform the center of projection
	//
	if(mode == PERSPECTIVE)
		makePerspectiveTransformation();
}

void setWindow(double _wxl, double _wxh, double _wyl, double _wyh)
{
	if(_wxl >= _wxh || _wyl >= _wyh)
		perror("bad window");
	wxl = _wxl;
	wxh = _wxh;
	wyl = _wyl;
	wyh = _wyh;
}

void setViewDepth(double _frontDistance, double _backDistance)
{
	if(frontDistance > backDistance)
		perror("Frontal Plane behind Back Plane");
		//
	frontDistance = _frontDistance;
	backDistance = _backDistance;
}

void setViewPort(double _vxl, double _vxh, double _vyl, double _vyh)
{
	if(_vxl >= _vxh || _vyl >= _vyh)
		perror("bad viewport");
	vxl = _vxl;
	vxh = _vxh;
	vyl = _vyl;
	vyh = _vyh;
}

void setMode(int _mode)
{
	mode = _mode;
}

void newView2()
{
	wsx = (vxh - vxl) / (wxh - wxl);
	wsy = (vyh - vyl) / (wyh - wyl);
}

void newView3()
{
	makeViewPlaneTransformation();
	newView2();
}

void setViewDistance(double _distance)
{
	distance = _distance;
}

void setPerspective(double x, double y, double z)
{
	center.x = x;
	center.y = y;
	center.z = z;
}

void clearBuffer()
{
	int i;
	DWORD *ptr = pxls;
	for(i = 0; i < WIDTH * HEIGHT; i++)
		*ptr++ = background;
}

void parallelTransform(Vector3d *p)
{
	p->x -= pc.x;
	p->y -= pc.y;
	p->z = 0;
}

void perspectiveTransform(Vector3d *p)
{
	double d = pc.z - p->z;
	if(fabs(d) < ROUNDOFF)
	{
		p->x = (p->x - pc.x) * VERYLARGE;
		p->y = (p->y - pc.y) * VERYLARGE;
		p->z = VERYLARGE;
	}
	else
	{
		p->x = (p->x * pc.z - pc.x * p->z) / d;
		p->y = (p->y * pc.z - pc.y * p->z) / d;
		p->z /= d;
	}
}

void enter(DWORD color)
{
	// Transform the point to camera space
	//
	viewPlaneTransform(&pen);
	//
	// Clipping against clipping planes
	//
	if(clipping && (pen.z > -frontDistance || pen.z < -backDistance))
		return;
	//
	if(pen.z < scale)
	{
		if(mode == PERSPECTIVE)
			perspectiveTransform(&pen);
		else
			parallelTransform(&pen);
		//
		// Clipping against frustum sides
		//
		if(pen.x >= wxl && pen.x <= wxh && pen.y >= wyl && pen.y <= wyh)
		{
			// Convert to screen coordinates
			//
			int xi = (int) (((pen.x - wxl) * wsx + vxl) * WIDTH + .5);
			int yi = (int) (((pen.y - wyl) * wsy + vyl) * HEIGHT + .5);
			//
			// Save point in output buffer
			//
			int pos = WIDTH - xi + (HEIGHT - yi - 1) * WIDTH;
			if(pos > 0 && pos < WIDTH * HEIGHT)
			{
				if(color == R || color == B)
					pxls[pos] = color;
				else if(pxls[pos] != R && pxls[pos] != G && pxls[pos] != B)
					pxls[pos] = color;
			}
		}
	}
}

void plot(double x, double y, double z, DWORD color)
{
	pen.x = x;
	pen.y = y;
	pen.z = z;
	enter(color);
}

double theta = 0;
BOOL rot = FALSE;

void setParallel(double dx, double dy, double dz)
{
	if((fabs(dx) + fabs(dy) + fabs(dz)) < ROUNDOFF)
		perror("No direction of projection");
}

/////////////////////////////////////////////////////
//                                                 //
//		Interface with the rest of the program     //
//                                                 //
/////////////////////////////////////////////////////

/*
 * Called by the main3d.c Windows interface.
 */
void initPlot(DWORD* pixels)
{
	pxls = pixels;
	initPalette();
	double h = sqrt(225) / (1.5 * GRID);
	position.x = (int)(10.0 / h);
	position.y = (int)(5.0 / h);
	position.z = (int)(10.0 / h);
	//
	direction.x = -position.x;		// camera axis
	direction.y = -position.y;
	direction.z = -position.z;
	normalize(&direction);
	//
	attitude.x = 0; 				// view-up direction
	attitude.y = 0;
	attitude.z = -1;
	//
	setViewPort(0.0, 1.0, 0.0, 1.0);
	setWindow(-0.5, 0.5, -0.5, 0.5);
	newView2();
	setViewDistance(0);
	setMode(PERSPECTIVE);
	if(mode == PERSPECTIVE)
	{
		double xc = position.x - direction.x * scale;
		double yc = position.y - direction.y * scale;
		double zc = position.z - direction.z * scale;
		setPerspective(xc, yc, zc);
	}
	newView3();
	setViewDepth(0, 10000);
	clearBuffer();
}

void Bresenham3D(DWORD color, int x1, int y1, int z1, const int x2, const int y2, const int z2)
{
    int i, dx, dy, dz, l, m, n, x_inc, y_inc, z_inc, err_1, err_2, dx2, dy2, dz2;
    int point[3];

    point[0] = x1;
    point[1] = y1;
    point[2] = z1;
    dx = x2 - x1;
    dy = y2 - y1;
    dz = z2 - z1;
    x_inc = (dx < 0) ? -1 : 1;
    l = abs(dx);
    y_inc = (dy < 0) ? -1 : 1;
    m = abs(dy);
    z_inc = (dz < 0) ? -1 : 1;
    n = abs(dz);
    dx2 = l << 1;
    dy2 = m << 1;
    dz2 = n << 1;

    if ((l >= m) && (l >= n))
    {
        err_1 = dy2 - l;
        err_2 = dz2 - l;
        for (i = 0; i < l; i++)
        {
        	plot(point[0], point[1], point[2], color);
            if (err_1 > 0)
            {
                point[1] += y_inc;
                err_1 -= dx2;
            }
            if (err_2 > 0)
            {
                point[2] += z_inc;
                err_2 -= dx2;
            }
            err_1 += dy2;
            err_2 += dz2;
            point[0] += x_inc;
        }
    }
    else if ((m >= l) && (m >= n))
    {
        err_1 = dx2 - m;
        err_2 = dz2 - m;
        for (i = 0; i < m; i++)
        {
        	plot(point[0], point[1], point[2], color);
            if (err_1 > 0)
            {
                point[0] += x_inc;
                err_1 -= dy2;
            }
            if (err_2 > 0)
            {
                point[2] += z_inc;
                err_2 -= dy2;
            }
            err_1 += dx2;
            err_2 += dz2;
            point[1] += y_inc;
        }
    }
    else
    {
        err_1 = dy2 - n;
        err_2 = dx2 - n;
        for (i = 0; i < n; i++)
        {
        	plot(point[0], point[1], point[2], color);
            if (err_1 > 0)
            {
                point[1] += y_inc;
                err_1 -= dz2;
            }
            if (err_2 > 0)
            {
                point[0] += x_inc;
                err_2 -= dz2;
            }
            err_1 += dy2;
            err_2 += dx2;
            point[2] += z_inc;
        }
    }
	plot(point[0], point[1], point[2], color);
}

/*
 * Called by automaton.c Loop3d thread
 */
void updatePlot()
{
	// Camera follows a circle and points to center
	//
	if(rot)
	{
		position.x = 1.5 * GRID * cos(theta);
		position.y = 1.5 * GRID * sin(theta);
		position.z = 0;
		//
		direction.x = -cos(theta);
		direction.y = -sin(theta);
		direction.z = 0;
		//
		attitude.x = 0;
		attitude.y = 0;
		attitude.z = 1;
		//
		theta += 0.01;
	}
	//
	// Create a new 2d view
	//
	newView2();
	//
	// Move the center of projection
	//
	if(mode == PERSPECTIVE)
	{
		setPerspective(position.x - scale * direction.x,
					   position.y - scale * direction.y,
					   position.z - scale * direction.z);
	}
	//
	// Create a new 3d view
	//
	newView3();
	//
	// Recreate pixels
	//
	clearBuffer();
	//
	int ndots = 128 / n;
	int i;
	for(i = 0; i < idx; i++)
	{
		Bresenham3D(colors[i], ndots*lxi[i], ndots*lyi[i], ndots*lzi[i], ndots*lxf[i], ndots*lyf[i], ndots*lzf[i]);
	}
	//
	if(showAxes)
	{
		Bresenham3D(X, 0, 0, 0, n * ndots / 2, 0, 0);
		Bresenham3D(Y, 0, 0, 0, 0, n * ndots / 2, 0);
		Bresenham3D(Z, 0, 0, 0, 0, 0, n * ndots / 2);
	}
}

void *DisplayLoop()
{
	while(true)
	{
		updatePlot();
		char *msg;
	 	asprintf(&msg, "%c %c %c %c %c %c %c %c", types[1] ? '1' : '_', types[2] ? '2' : '_', types[3] ? '3' : '_', types[4] ? '4' : '_', types[5] ? '5' : '_', types[6] ? '6' : '_', types[7] ? '7' : '_', types[8] ? '8' : '_');
		vprints(20, 30, msg);
		SetDIBits(myCompatibleDC, myBitmap, 0, HEIGHT, pixels, &bmInfo, 0);
		BitBlt(hdc, 0, 0, WIDTH, HEIGHT, myCompatibleDC, 0, 0, SRCCOPY);
		delay(100);
	}
	return NULL;
}
